unit Environ;

{ Environment variables.
  To use this unit, simply add Environ to the uses statement,
  and the Env variable is automatically initialized.

  Env['PATH'] returns a string that is the value of the
  PATH environment variable.  For undefined variables, an
  empty string is returned.  Assign to Env[...] to change
  the value of an environment variable or to create a new one.
  Call Delete or Remove to delete a variable.

  If you want to use the environment in a LoadModule call,
  call the GetEnvBlock function, which returns an EnvBlock pointer
  for the environment.  If you modified the environment variables,
  then a new EnvBlock is allocated; otherwise, it uses the same
  EnvBlock as the current application.  If you do not modify the
  environment, and the environment appears to be the global
  Windows environment, then EnvBlock returns 0, which tells
  LoadModule to use the global Windows enviornment.

  In Win32, GetEnvBlock always returns nil to use the current
  process environment.

  The caller is responsible for freeing the EnvBlock (with GlobalFree),
  since it is not safe to free the EnvBlock until after the child process
  is done using it, and only the caller can know when that is.  If the
  caller knows that the child is done with it, call the FreeEnvBlock procedure,
  which also handles the case where the EnvBlock is the global DOS environment
  and cannot be freed, which is indicated by an EnvBlock of zero.

  To try to be safe, TEnvironment copies strings from the environment
  for its own private use.  Thus, if the parent accidentally frees
  the environment, the child tries to grab all its information initially,
  and does not refer to the EnvBlock again until the program calls the
  EnvBlock method, at which time, it checks to see if the pointer is valid.

  The Env object represents the global environment in Windows 3.x,
  or the process environment in Win32. You can create other TEnvironment
  objects. The AllocEnvBlock function returns a newly allocated environment
  that represents the TEnvironment variables and values. You are
  responsible for freeing that environment block when you are finished
  with it by calling FreeEnvBlock.
}

interface

  uses Classes;

  type
    TEnvBlock = {$ifdef WIN32} PChar {$else} Word {$endif};
    TEnvironment = class(TPersistent)
    private
      fModified: Boolean;
      fStrings: TStringList;
      fEnvSize: Cardinal;
    {$ifndef WIN32}
      fEnvBlock: TEnvBlock;
    {$endif}
      function GetEnv(const EnvVar: string): string;
      procedure SetEnv(const EnvVar: string; const Value: string);
      function GetCount: Integer;
      function GetName(Index: Integer): string;
      function GetSorted: Boolean;
      procedure SetSorted(Sorted: Boolean);
    protected
      constructor CreateFromEnv(Env: PChar); virtual;
      procedure FreeStrings; virtual;
      procedure Initialize(Env: PChar); virtual;
      function GetValuePtr(Index: Integer): PString;
      procedure SetValuePtr(Index: Integer; Ptr: PString);
    public
      constructor Create; virtual;
      destructor Destroy; override;
      function AllocEnvBlock: TEnvBlock; virtual;
      procedure Assign(Source: TPersistent); override;
      property Count: Integer read GetCount;
      property Env[const EnvVar: string]: string read GetEnv write SetEnv; default;
      property Sorted: Boolean read GetSorted write SetSorted;
      property Names[Index: Integer]: string read GetName;
      procedure Delete(Index: Integer); virtual;
      procedure Remove(const EnvVar: string);
      property Modified: Boolean read fModified;
      property Size: Cardinal read fEnvSize;
    end;

  { Global environment }
  var
    Env: TEnvironment;

  { Return an EnvBlock that can be used in a LoadModule, or similar call.
    If the environment has not been modified, return a reference to the
    global environment block. }
  function GetEnvBlock: TEnvBlock;
  { Free the environment block that was returned by GetEnvBlock, unless
    that is the global environment block. Every call to GetEnvBlock must
    have a corresponding FreeEnvBlock. }
  procedure FreeEnvBlock(EnvBlock: TEnvBlock);

implementation

uses SysUtils, WinTypes, WinProcs;

{ Delphi 2 defines some simple procedures for working with strings.
  Define similar procedures in Delphi 1, for portability. }
{$ifndef WIN32}
{ Set the length of string Str to Length. }
procedure SetLength(var Str: string; Length: Byte);
begin
  Str[0] := Chr(Length)
end;

{ Set a Pascal string, Str, to a PChar, Data, of length, Length. }
procedure SetString(var Str: string; Data: PChar; Length: Byte);
begin
  Str[0] := Chr(Length);
  Move(Data^, Str[1], Length);
end;
{$endif}

{ Return the environment block for use in a LoadModule, CreateProcess
  or similar call. If that is the global environment, then return
  0 or nil. Otherwise, return a reference to the environment block. }
function GetEnvBlock: TEnvBlock;
begin
{$ifdef WIN32}
  Result := nil;
{$else}
  Result := Env.AllocEnvBlock;
{$endif}
end;

procedure FreeEnvBlock(EnvBlock: TEnvBlock);
begin
{$ifdef WIN32}
  if EnvBlock <> nil then
    HeapFree(GetProcessHeap, 0, EnvBlock);
{$else}
  if EnvBlock <> 0 then
    if GlobalSize(EnvBlock) > 0 then
    begin
      GlobalFree(EnvBlock);
      if Env.fEnvBlock = EnvBlock then
        Env.fEnvBlock := 0;
    end;
{$endif}
end;

{ Create a new, empty TEnvironment. }
constructor TEnvironment.Create;
begin
  fStrings := TStringList.Create;
  inherited Create;
end;

{ Create a new TEnvironment, reading variables & values from Env. }
constructor TEnvironment.CreateFromEnv(Env: PChar);
begin
  Create;
  Initialize(Env);
{$ifndef WIN32}
  if LongRec(Env).Lo = 0 then
    fEnvBlock := LongRec(Env).Hi;
{$endif}
end;

procedure TEnvironment.Assign(Source: TPersistent);
var
  I: Integer;
begin
  if Source is TEnvironment then
  begin
    fStrings.Assign(TEnvironment(Source).fStrings);
    fModified := True;
  {$ifndef WIN32}
    fEnvBlock := 0;
  {$endif}
    fEnvSize := 1;  { trailing #0 }
    with fStrings do
      for I := 0 to Count-1 do
        fEnvSize := fEnvSize + Length(Strings[I]) + Length(GetValuePtr(I)^) + 2;
  end
  else
    inherited Assign(Source);
end;

{ Destroy the TEnvironment, and free all the memory }
destructor TEnvironment.Destroy;
begin
  FreeStrings;
  fStrings.Free;
  inherited Destroy;
end;

{ Return whether the environment is automatically sorted. Note that
  GetSorted might return False but the environment names could be
  coincidentally sorted. }
function TEnvironment.GetSorted: Boolean;
begin
  Result := fStrings.Sorted;
end;

{ The user can set the environment to be sorted. I don't see this
  as being particularly useful, but there is no reason to hide
  this functionality. }
procedure TEnvironment.SetSorted(Sorted: Boolean);
begin
  if Sorted <> fStrings.Sorted then
  begin
    fModified := True;
  {$ifndef WIN32}
    fEnvBlock := 0;
  {$endif}
  end;
  fStrings.Sorted := Sorted;
end;

{ Free the memory for all the environment values }
procedure TEnvironment.FreeStrings;
var
  I: Integer;
begin
  with fStrings do
    for I := 0 to Count-1 do
    begin
      DisposeStr(GetValuePtr(I));
      SetValuePtr(I, nil);
    end;
end;

{ Initialize the environment from a string. The string has the form:
    var=value#0var=value#0#0
  Each variable is expressed as a var=value pair, separated by an
  equal sign and terminated by a zero byte.  After the last variable's
  value, there is an extra zero byte. }
procedure TEnvironment.Initialize(Env: PChar);
var
  Name: string;
  ValStr: PString;
  Value: PChar;
begin
  fStrings.Clear;
  fEnvSize := 1;        { for the terminating zero byte }

  while Env[0] <> #0 do
  begin
    { Find the value, which comes after the equal sign. }
    Value := StrScan(Env, '=');
    if Value = nil then
    begin
      { This should never happen, but it's better to be safe. }
      Value := StrEnd(Env);
      Name := StrPas(Env);
    end
    else
    begin
      SetString(Name, Env, Value-Env);
      Inc(Value);
    end;
    ValStr := NewStr(StrPas(Value));
    if Name <> '' then
    begin
      { Win32 uses environment variables without names for passing
        initial directories to a child process. Ignore them. }
      fStrings.AddObject(Name, TObject(ValStr) );
      fEnvSize := fEnvSize + Length(Name) + Length(ValStr^) + 2;
    end;

    { Skip to the end of the value, and skip over its #0 byte }
    Env := StrEnd(Value);
    Inc(Env);
  end
end;

{ Allocate a EnvBlock for the environment, and copy all the strings into it }
function TEnvironment.AllocEnvBlock: TEnvBlock;
var
  I: Integer;
  Ptr: PChar;
begin
{$ifdef WIN32}
  Result := HeapAlloc(GetProcessHeap, 0, Size);
  if Result = nil then
    OutOfMemoryError;
  Ptr := Result;
{$else}
  if fEnvBlock <> 0 then
  begin
    Result := fEnvBlock;
    Exit;
  end;

  Result := GlobalAlloc(Gmem_Fixed or Gmem_Share, Size);
  if Result = 0 then
    OutOfMemoryError;
  Ptr := GlobalLock(Result);
  if Ptr = nil then
  begin
    GlobalFree(Result);
    OutOfMemoryError;
  end;
{$endif}

  try
    for I := 0 to Count-1 do
    begin
      StrPCopy(Ptr, fStrings[I]);
      Inc(Ptr, Length(fStrings[I]));
      Ptr[0] := '=';
      Inc(Ptr);
      StrPCopy(Ptr, GetValuePtr(I)^);
      Inc(Ptr, Length(GetValuePtr(I)^) + 1);
    end;
    Ptr[0] := #0;
  except
  {$ifdef WIN32}
    HeapFree(GetProcessHeap, 0, Result);
  {$else}
    GlobalUnlock(Result);
    GlobalFree(Result);
  {$endif}
    raise;
  end;
{$ifndef WIN32}
  fEnvBlock := Result;
{$endif}
end;

{ Return the number of environment variables defined }
function TEnvironment.GetCount: Integer;
begin
  Result := fStrings.Count
end;

{ Get the name of the environment variable at Index }
function TEnvironment.GetName(Index: Integer): string;
begin
  Result := fStrings.Strings[Index]
end;

{ Get the value of the environment variable named EnvVar }
function TEnvironment.GetEnv(const EnvVar: string): string;
var
  Index: Integer;
begin
  Index := fStrings.IndexOf(EnvVar);
  if Index < 0 then
    Result := ''
  else
    Result := GetValuePtr(Index)^;
end;

{ Return a pointer to the value string at Index }
function TEnvironment.GetValuePtr(Index: Integer): PString;
begin
  Result := PString(fStrings.Objects[Index]);
end;

{ Store a string pointer value }
procedure TEnvironment.SetValuePtr(Index: Integer; Ptr: PString);
begin
  with fStrings do
  begin
    if (Ptr <> nil) and (Objects[Index] <> nil) and (PString(Objects[Index])^ <> Ptr^) then
    begin
      fModified := True;
    {$ifndef WIN32}
      fEnvBlock := 0;
    {$endif}
    end;
    Objects[Index] := TObject(Ptr);
  end;
end;

{ Delete the environment variable at Index. Release the EnvBlock now
  because we know we modified it, but we don't need to create a new
  one just yet. }
procedure TEnvironment.Delete(Index: Integer);
begin
  if (Index >= 0) and (Index < fStrings.Count) then
  begin
  {$ifdef WIN32}
    if Self = Environ.Env then
      SetEnvironmentVariable(PChar(fStrings[Index]), nil);
  {$endif}
    fEnvSize := fEnvSize - Length(fStrings[Index]) - Length(GetValuePtr(Index)^) - 2;
    fStrings.Delete(Index);
    fModified := True;
  {$ifndef WIN32}
    fEnvBlock := 0;
  {$endif}
  end;
end;

{ Delete the environment variable named, EnvVar }
procedure TEnvironment.Remove(const EnvVar: string);
var
  Index: Integer;
begin
  Index := fStrings.IndexOf(EnvVar);
  if Index >= 0 then
    Delete(Index);
end;

{ Set an environment variable.  If it is a new one, then add the new
  variable to the list.  If the variable already exists, then just
  replace its value with the new value. Release the EnvBlock because
  we modified it, but we don't need to allocate a new one yet. }
procedure TEnvironment.SetEnv(const EnvVar: string; const Value: string);
var
  Index: Integer;
  OldVal: PString;
begin
  Index := fStrings.IndexOf(EnvVar);
  if Index < 0 then
  begin
    { New environment variable }
    fStrings.AddObject(EnvVar, TObject(NewStr(Value)));
    fEnvSize := fEnvSize + Length(EnvVar) + Length(Value) + 2;
  end
  else if GetValuePtr(Index)^ = Value then
    { Same name, same value }
    Exit
  else
  begin
    { new value for an existing variable }
    OldVal := GetValuePtr(Index);
    fStrings.Objects[Index] := nil;
    fEnvSize := fEnvSize - Length(OldVal^);
    DisposeStr(OldVal);
    SetValuePtr(Index, NewStr(Value));
    fEnvSize := fEnvSize + Length(Value);
  end;
  fModified := True;
{$ifdef WIN32}
  if Self = Environ.Env then
    SetEnvironmentVariable(PChar(EnvVar), PChar(Value));
{$else}
  fEnvBlock := 0;
{$endif}
end;

{ When the application terminates, free the environment. }
procedure Terminate; far;
begin
  Env.Free
end;

initialization
  AddExitProc(Terminate);
{$ifdef WIN32}
  Env := TEnvironment.CreateFromEnv(GetEnvironmentStrings);
{$else}
  Env := TEnvironment.CreateFromEnv(GetDOSEnvironment);
{$endif}
end.
{--=>revision-history<=--}
{1 ENVIRON.PAS 3-Sep-95,23:57:30,`lisch' CHANGE=INITIAL                       }
{     Initial version.                                                        }
{--=>revision-history<=--}
