unit Launch;

{ Launch an application and wait until it completes. Then return.
  It is possible to write the same code that calls LoadModule for
  Windows 3.1 or Win32. The problem lies in detecting when the program
  finishes. Win32 has convenient Wait functions, but only if you have
  the process ID. In Windows 3.1, you need to check the module usage,
  but that doesn't work in Win32.
}

interface

uses WinTypes, WinProcs, SysUtils, Environ;

type
  TShow = array[0..1] of Word;
  PShow = ^TShow;
  TLoadParams = record
    segEnv: TEnvBlock;
    lpszCmdLine: PChar;
    lpShow: PShow;
    lpReserved: Pointer;
  end;

procedure LaunchApp(const AppPath, CmdLine, StartDir: string; Env: TEnvBlock);

implementation

{ Allocate a PChar and copy a Pascal string to it.
  The caller must free the string with StrFree. Delphi 2
  and later has builtin support for converting a Pascal
  string to PChar, so type casting works without memory allocation. }
function StrMake(const PasStr: string): PChar;
begin
{$ifdef WIN32}
  Result := PChar(PasStr);
{$else}
  Result := StrAlloc(Length(PasStr)+1);
  StrPCopy(Result, PasStr);
{$endif}
end;

procedure StrFree(Str: PChar);
begin
{$ifndef WIN32}
  StrDispose(Str);
{$endif}
end;

procedure LaunchApp(const AppPath, CmdLine, StartDir: string; Env: TEnvBlock);
var
{$ifdef WIN32}
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  FullCmdLine: string;
{$else}
  LoadParams: TLoadParams;  { LoadModule parameters }
  Show: TShow;
  lpAppPath: PChar;
  lpCmdLine: PChar;
  hModule: THandle;
{$endif}
begin
{$ifdef WIN32}
  FullCmdLine := AppPath + ' ' + CmdLine;
  FillChar(StartupInfo, SizeOf(StartupInfo), 0);
  StartupInfo.cb := SizeOf(StartupInfo);
  Win32Check(CreateProcess(nil, PChar(FullCmdLine), nil, nil, True, 0, Env,
                       PChar(StartDir), StartupInfo, ProcessInfo));

  WaitForSingleObject(ProcessInfo.hProcess, Infinite);

{$else}
  ChDir(StartDir);
  { Convert strings to zero-terminated character arrays. }
  lpAppPath := StrMake(AppPath);
  lpCmdLine := StrMake(' ' + CmdLine); { Windows concatenates AppPath and CmdLine,
                                         so add a space character between them. }

  { Setup the LoadModule parameters. }
  LoadParams.segEnv := GetEnvBlock;
  LoadParams.lpszCmdLine := lpCmdLine;
  Show[0] := 2;
  Show[1] := Sw_Normal;
  LoadParams.lpShow := @Show;
  LoadParams.lpReserved := nil;

  hModule := LoadModule(lpAppPath, @LoadParams);
  StrFree(lpAppPath);
  StrFree(lpCmdLine);

  if hModule < 32 then
    raise Exception.CreateFmt('Cannot run %s %s', [AppPath, CmdLine]);

  { Wait for program to finish. }
  while GetModuleUsage(hModule) > 0 do
    Yield;
{$endif}
end;

end.
